/* use select to read from 1 file descriptor (and wait on kbd input for stop)
 * based on Serial Programming Guide for POSIX Operating Systems
 */

#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include "rs22812.h"

#define BAUDRATE    B4800
#define MODEMDEVICE "/dev/ttyS1"
#define _POSIX_SOURCE 1 /* POSIX compliant source */
#define QUIET   0
#define VERBOSE 1
#define MAXLOOPS 500

int setup_raw_fd(int fd, struct termios* poldtio)
/* assumes fd has already been opened */
{
  int c, res;
  struct termios newtio;
  char buf[255];

  tcgetattr(fd, poldtio); /* save current port settings */

  bzero(&newtio, sizeof(newtio));
  newtio.c_cflag = BAUDRATE | CRTSCTS | CS8 | CLOCAL | CREAD;
  newtio.c_iflag = IGNPAR;
  newtio.c_oflag = 0;

  /* set input mode (non-canonical, no echo,...) */
  newtio.c_lflag = 0;

  newtio.c_cc[VTIME]    = 0;   /* inter-character timer unused */
  newtio.c_cc[VMIN]     = 9/*5*/;   /* blocking read until 5 chars received */

  tcflush(fd, TCIFLUSH);
  tcsetattr(fd,TCSANOW,&newtio);
  return fd;
}

int setup_port(char* devicename, struct termios* poldtio)
{
  int fd;

  /* printf("\t* about to open %s...\r\n", devicename); */
  fd = open(devicename, O_RDWR | O_NOCTTY );
  if (fd <0) {perror(devicename); exit(-1); }

  return setup_raw_fd(fd, poldtio);
}

int cleanup_port(int fd, struct termios* poldtio)
{
  tcsetattr(fd, TCSANOW, poldtio);
}

int dumpbytes(int fd, int count)
{
	int i;
	char* buf;
	int rrv;

	/* printf("\t* dumpbytes - fd: %d, count: %d *\r\n", fd, count); */
	buf = (char*)malloc(count+1);
	if (buf>=0)
	{
		bzero(buf, count+1);
		rrv = read(fd, buf, count);
		for (i=0; i<count; i++)
		{
			printf("%2x ", buf[i] & 0xff);
		}
		free(buf);
		printf("\n\r");
	}
}

int nbytes(int fd, int verbose)
{
	int bytes;

	ioctl(fd, FIONREAD, &bytes);
	if (verbose)
		printf("\r\n\t* fd %d had %d bytes *\r\n", fd, bytes);
	return bytes;
}

void process_fd(int fd)
{
#define DBUFSZ 255
	char  dbuf[DBUFSZ]; /* display buffer */
	char* rbuf;         /* read    buffer */
	int   n;
	int   rrv;
	RS22812_PACKET* pak;

	n = nbytes(fd, QUIET);
	if (n < 0)
		printf("\t* process_fd error: negative count! *\r\n");
	else
	{
		rbuf = (char*)malloc(n+1);
		if (rbuf > 0)
		{
			bzero(rbuf, n+1);
			rrv = read(fd, rbuf, n);
		}
		if (rrv > 0)
		{
			bzero(dbuf, DBUFSZ);
			pak = rs22812_packet_new();
			rs22812_packet_read(pak, rbuf);
			rs22812_packet_cook(pak);
			/*rs22812_packet_dump(pak);*/
			rs22812_packet_sprint(pak, dbuf, DBUFSZ);
			printf("%s", dbuf);
			rs22812_packet_delete(pak);
		}
	}
}

int do_read(int fd)
{
	int n;
	int socket = 0;
	int max_fd;
	int stop   = 0;

	fd_set input;
	struct timeval timeout;

	/* Initialize the input set */
	FD_ZERO(&input);
	FD_SET(fd, &input);
	FD_SET(socket, &input);
	max_fd = (socket > fd ? socket : fd) + 1;
	/* printf("\r\n\t *do_read - max_fd = %d\r\n", max_fd); */

	/* Initialize the timeout structure */
	timeout.tv_sec = 10;
	timeout.tv_usec = 0;
	/* Do the select */
	n = select(max_fd, &input, NULL, NULL, &timeout);
	/* See if there was an error */
	if (n < 0)
		perror("select failed");
	else if (n == 0)
		puts("\n\r\t* TIMEOUT *\n\r");
	else
	{
		/* We have input */
		if (FD_ISSET(fd, &input))
			process_fd(fd);
		if (FD_ISSET(socket, &input))
			stop = 1;
	}

	return stop;
}

int main(void)
{
	struct termios oldtio;
	struct termios coldtio;
	int fd;
	int stop = 0;
	int loops = 0;
	int maxloops = MAXLOOPS;

	fd = setup_port(MODEMDEVICE, &oldtio);
	/*setup_raw_fd(0, &coldtio);*/

	for (loops = 0, stop=0; loops < maxloops; loops++)
	{
		stop = do_read(fd);
		if (stop)
	break;
	}

	/* cleanup_port(0,  &coldtio); */
	cleanup_port(fd, &oldtio);

	printf("stopped %s%s\n", stop ? "for user " : "", 
	 	    loops >= maxloops ? "for max loops" : "" );

	return 0;
}
